from math import sqrt, asin, isnan, nan


def __circleIntegral(x0, r, x):
    return (x - x0) * sqrt(round(r ** 2 - (x - x0) ** 2, 3)) + r ** 2 * asin(round((x - x0) / r, 3))


def __getCirclesArea(x0, r, a, b):
    return __circleIntegral(x0, r, b) - __circleIntegral(x0, r, a)


def __getArea(size_length, size_height, x0, y0, r, indent_length, indent_height):
    x0 -= indent_length
    y0 -= indent_height
    h = y0 if y0 - r < 0 else ((size_height - y0) if y0 + r > size_height else nan)
    local_s = __getCirclesArea(x0, r, max(x0 - r, 0), min(x0 + r, size_length))
    if not isnan(h):
        d = sqrt(r ** 2 - h ** 2)
        x1, x2 = max(x0 - d, 0), min(x0 + d, size_length + 2 * size_length)
        local_s -= (__getCirclesArea(x0, r, x1, x2) - 4 * h * d) / 2
    return local_s



from math import sqrt

from . import __chess, __inline, __random


def __convert(geometry_data):
    geometry_data['size'] = float(geometry_data['size'][0]), float(geometry_data['size'][1])
    geometry_data['porosityNumber'] = float(geometry_data['porosityNumber'])
    geometry_data['radius'] = float(geometry_data['radius'][0]), float(geometry_data['radius'][1])
    geometry_data['minimumDistance'] = float(geometry_data['minimumDistance'])
    geometry_data['periodicity'] = geometry_data['periodicity'][:2]
    geometry_data['indent'] = float(geometry_data['indent'][0]), float(geometry_data['indent'][1])


def __calculate_lines_indents(size_length, size_height, obstacles, periodicity_length, periodicity_height):
    lines_x = round(sqrt(obstacles * size_length / size_height))
    lines_y = round(sqrt(obstacles * size_height / size_length))
    while (lines_x - 1) * lines_y >= obstacles:
        lines_x -= 1
    while lines_x * lines_y < obstacles:
        if lines_x * (lines_y + 1) - obstacles > lines_y:
            lines_x += 1
        else:
            lines_y += 1
    bottom = lines_x + (1 if not periodicity_length else -1)
    if bottom != 0:
        lines_indent_x = size_length / bottom
    else:
        lines_indent_x = size_length
    bottom = lines_y + (1 if not periodicity_height else -1)
    if bottom != 0:
        lines_indent_y = size_height / bottom
    else:
        lines_indent_y = size_height
    return lines_x, lines_y, lines_indent_x, lines_indent_y


def __generate(geometry_data):

    __convert(geometry_data)

    if geometry_data['order'] == 0:
        generator = __random

    elif geometry_data['order'] == 1:
        generator = __chess
        geometry_data['settings'] = __calculate_lines_indents(*geometry_data['size'], 2 * geometry_data['porosityNumber'] - 1, *geometry_data['periodicity'])

        del geometry_data['porosityNumberButton']

    elif geometry_data['order'] == 2:
        generator = __inline
        geometry_data['settings'] = __calculate_lines_indents(*geometry_data['size'], geometry_data['porosityNumber'], *geometry_data['periodicity'])

        del geometry_data['porosityNumberButton']

    else:
        raise TypeError

    del geometry_data['order']

    disks, porosity, time = generator.__generate(**geometry_data)

    return disks, porosity, time



from json import load, dump
from ctypes import CDLL, POINTER, CFUNCTYPE, c_void_p, c_longdouble, c_int

from .Classes import Domain, Obstacle


def loadLibrary(filename: str) -> CDLL:
    '''
    Loads shared library with random algorithm written in c++.

    filename: str - path to library

    return - ctypes.CDLL
    '''
    lib = CDLL(filename)
    lib.generate.argtypes = (Domain, CALLBACK_TYPE)
    lib.generate.restype = POINTER(c_longdouble * 2)
    return lib


def loadGeometry(filename: str) -> dict:
    '''
    Loads dict with domain settings from file.

    filename: str - path to .json file with domain settings

    return - Domain
    '''
    with open(filename) as file:
        data = load(file)
    return data


def saveDomain(filename: str, domain: Domain) -> None:
    if not filename.endswith('.json'):
        filename += '.json'
    
    with open(filename, 'w') as file:
        dump(domain.toJSON())


def saveObstacles(filename: str, obstacles: tuple[Obstacle]) -> None:
    if not filename.endswith('.json'):
        filename += '.json'
    
    with open(filename, 'w') as file:
        dump([[obstacle.center.x, obstacle.center.y, obstacle.center.z, obstacle.radius] for obstacle in obstacles], file, indent=4)


CALLBACK_TYPE = CFUNCTYPE(c_void_p, c_int, c_longdouble)



from math import sqrt, asin, isnan, pi, nan


def __round_neg(x):
    if x < 0:
        return -round(x)
    else:
        return x


def __circle_integral(x0, r, x):
    return (x - x0) * sqrt(__round_neg(r ** 2 - (x - x0) ** 2)) + r ** 2 * asin(round((x - x0) / r, 3))


def __get_circles_area(x0, r, a, b):
    return __circle_integral(x0, r, b) - __circle_integral(x0, r, a)


def __get_area(size_length, size_height, x0, y0, r, indent_length, indent_height):
    x0 -= indent_length
    y0 -= indent_height
    h = y0 if y0 - r < 0 else ((size_height - y0) if y0 + r > size_height else nan)
    local_s = __get_circles_area(x0, r, max(x0 - r, 0), min(x0 + r, size_length))
    if not isnan(h):
        d = sqrt(__round_neg(r ** 2 - h ** 2))
        x1, x2 = max(x0 - d, 0), min(x0 + d, size_length + 2 * size_length)
        local_s -= (__get_circles_area(x0, r, x1, x2) - 4 * h * d) / 2
    return local_s


def __get_volume(size_length, size_height, size_width, x0, y0, z0, r, indent_length, indent_height, indent_width):
    n = 1000
    z0 -= indent_width
    v = 0
    if z0 < 0:
        a = r + z0
        d = a / n
        for i in range(n + 1):
            h = sqrt(__round_neg(r ** 2 - (i * d) ** 2))
            v += d * __get_area(size_length, size_height, x0, y0, h, indent_length, indent_height)
    elif z0 > size_width:
        a = r - z0 + size_width
        d = a / n
        for i in range(n + 1):
            h = sqrt(__round_neg(r ** 2 - (i * d) ** 2))
            v += d * __get_area(size_length, size_height, x0, y0, h, indent_length, indent_height)
    else:
        a = (r if z0 - r > 0 else z0)
        d = a / n
        for i in range(1, n + 1):
            h = sqrt(__round_neg(r ** 2 - (i * d) ** 2))
            v += d * pi * h ** 2  # get_area(size_length, size_height, x0, y0, h, indent_length, indent_height)
        a = (r if z0 + r < size_width else r - size_width + z0)
        d = a / n
        for i in range(1, n + 1):
            h = sqrt(__round_neg(r ** 2 - (i * d) ** 2))
            v += d * pi * h ** 2  # get_area(size_length, size_height, x0, y0, h, indent_length, indent_height)
    return v



from time import perf_counter
from random import random

from . import __area


def __checkForIntersection(x, y, r, disks, minimumDistance):
    for _x, _y, _, _r in disks:
        if (_x - x) ** 2 + (_y - y) ** 2 <= (_r + r + minimumDistance) ** 2:
            return True
    return False


def __generate(size, periodicity, indent, porosityNumberButton, porosityNumber, radius, minimumDistance):
    indent_length = 0 if periodicity[0] else indent[0]
    indent_width = 0 if periodicity[1] else indent[1]

    disks = set()

    total_area = size[0] * size[1]
    current_area = float(total_area)
    current_porosity = 100

    running = True
    start = perf_counter()

    while running:
        if len(disks) >= porosityNumber and porosityNumberButton == 1 or current_porosity <= porosityNumber and porosityNumberButton == 0:
            running = False

        x = indent_length + random() * size[0]
        y = indent_width + random() * size[1]
        r = radius[0] + random() * (radius[1] - radius[0])
    
        if not __checkForIntersection(x, y, r, disks, minimumDistance):
            need_x = 2 if (x - r < 0 or x + r > size[0] + indent_length) and periodicity[0] else 0
            need_y = 2 if (y - r < 0 or y + r > size[1] + indent_width) and periodicity[1] else 0
    
            if len(disks) + need_x + need_y >= porosityNumber and porosityNumberButton == 1:
                continue
    
            copies = []
            if need_x:
                if x - r < 0:
                    _x = x + size[0]
                elif x + r > size[0] + indent_length:
                    _x = x - size[0]
                copies.append((_x, y))
            if need_y:
                if y - r < 0:
                    _y = y + size[1]
                elif y + r > size[1] + indent_width:
                    _y = y - size[1]
                copies.append((x, _y))
            if need_x and need_y:
                copies.append((_x, _y))
    
            for _x, _y in copies:
                if __checkForIntersection(_x, _y, r, disks, minimumDistance):
                    break
            else:
                copies.append((x, y))
                for _x, _y in copies:
                    disks.add((_x, _y, 0, r))
                    current_area -= __area.__getArea(size[0], size[1], _x, _y, r, indent_length, indent_width)
                # disks.add((x, y, 0, r))
                # current_area -= __area.__getArea(size[0], size[1], x, y, r, indent_length, indent_width)
                current_porosity = 100 * (current_area / total_area)

    running = False
    end = perf_counter()

    return disks, current_porosity, end - start



from random import random
from time import perf_counter

from . import __area


def __generate(size, periodicity, indent, porosityNumber, radius, minimumDistance, settings):
    indent_length = 0 if periodicity[0] else indent[0]
    indent_width = 0 if periodicity[1] else indent[1]
    indent_height = 0 if periodicity[2] else indent[2]

    indent_x = (indent_length + settings[3]) if not periodicity[0] else 0
    indent_y = (indent_width + settings[4]) if not periodicity[1] else 0
    indent_z = (indent_height + settings[5]) if not periodicity[2] else 0

    maximum_radius = min(min(settings[2] ** 2, settings[3] ** 2) - radius[0] - minimumDistance, radius[1])
    minimum_radius = maximum_radius if maximum_radius < radius[0] else radius[0]

    spheres = set()
    spheres3d = [[[None for _ in range(settings[2])] for _ in range(settings[1])] for _ in range(settings[0])]

    total_volume = size[0] * size[1] * size[2]
    current_volume = float(total_volume)
    current_porosity = 0

    start = perf_counter()
    try:
        for i in range(settings[0]):
            if i == 0 and periodicity[0]:
                periodicity_x = settings[0] - 1
            else:
                periodicity_x = None
            for j in range(settings[1]):
                if j == 0 and periodicity[1]:
                    periodicity_y = settings[1] - 1
                else:
                    periodicity_y = None
                for k in range(settings[2]):
                    if k == 0 and periodicity[2]:
                        periodicity_z = settings[2] - 1
                    else:
                        periodicity_z = None
                    if spheres3d[i][j][k] is None:
                        x, y, z = indent_x + settings[3] * i, indent_y + settings[4] * j, indent_z + settings[5] * k
                        max_r = maximum_radius
                        for n, m, t in ((1, 0, 0), (-1, 0, 0), (0, 1, 0), (0, -1, 0), (0, 0, 1), (0, 0, -1)):
                            r, d, c = i + n, j + m, k + t
                            if 0 <= r < settings[0] and 0 <= d < settings[1] and 0 <= c < settings[2]:
                                if spheres3d[r][d][c] is not None:
                                    if n != 0:
                                        max_rt = abs(spheres3d[r][d][c][0] - x) - spheres3d[r][d][c][3] - minimumDistance
                                    elif m != 0:
                                        max_rt = abs(spheres3d[r][d][c][1] - y) - spheres3d[r][d][c][3] - minimumDistance
                                    else:
                                        max_rt = abs(spheres3d[r][d][c][2] - z) - spheres3d[r][d][c][3] - minimumDistance
                                    max_r = min(max_rt, max_r)
                        spheres3d[i][j][k] = (x, y, z, minimum_radius + random() * (max_r - minimum_radius))
                        if periodicity_x is not None:
                            if spheres3d[periodicity_x][j][k] is None:
                                spheres3d[periodicity_x][j][k] = (indent_x + settings[3] * periodicity_x,
                                                             y,
                                                             z,
                                                             spheres3d[i][j][k][3])
                        if periodicity_y is not None:
                            if spheres3d[i][periodicity_y][k] is None:
                                spheres3d[i][periodicity_y][k] = (x,
                                                             indent_y + settings[4] * periodicity_y,
                                                             z,
                                                             spheres3d[i][j][k][3])
                        if periodicity_z is not None:
                            if spheres3d[i][j][periodicity_z] is None:
                                spheres3d[i][j][periodicity_z] = (x,
                                                             y,
                                                             indent_z + settings[5] * periodicity_z,
                                                             spheres3d[i][j][k][3])
                    spheres.add(tuple(spheres3d[i][j][k]))
                    current_volume -= __area.__get_volume(size[0], size[1], size[2], *spheres3d[i][j][k], indent_length, indent_height, indent_width)
                    current_porosity = 100 * (current_volume / total_volume)
                    progress = (len(spheres), current_porosity)
                    if len(spheres) >= porosityNumber:
                        raise KeyboardInterrupt

    except KeyboardInterrupt:
        ...
    
    end = perf_counter()

    return spheres, current_porosity, end - start



from ctypes import (
    Structure,
    c_longdouble,
    c_bool,
    c_short,
    c_uint,
    c_int,
    c_bool,
)


class Point(Structure):
    _fields_ = [
        ('x', c_longdouble),
        ('y', c_longdouble),
        ('z', c_longdouble),
    ]

    def __repr__(self) -> str:
        return f'({self.x}, {self.y}, {self.z})'
    
    def __call__(self) -> tuple:
        return self.x, self.y, self.z
    
    def __getitem__(self, key) -> int:
        if key == 0:
            return self.x
        elif key == 1:
            return self.y
        elif key == 2:
            return self.z
        else:
            raise KeyError('Uncorrect key')


class Periodicity(Structure):
    _fields_ = [
        ('x', c_bool),
        ('y', c_bool),
        ('z', c_bool),
    ]
    
    def __call__(self) -> tuple:
        return self.x, self.y, self.z
    
    def __getitem__(self, key) -> int:
        if key == 0:
            return self.x
        elif key == 1:
            return self.y
        elif key == 2:
            return self.z
        else:
            raise KeyError('Uncorrect key')


class Radius(Structure):
    _fields_ = [
        ('min', c_longdouble),
        ('max', c_longdouble),
    ]
    
    def __call__(self) -> tuple:
        return self.min, self.max


class Counter(Structure):
    _fields_ = [
        ('type', c_short),
        ('number', c_uint),
        ('porosity', c_longdouble),
    ]
    
    def __call__(self) -> tuple:
        return self.type, self.number, self.porosity


class Domain(Structure):
    _fields_ = [
        ('dimension', c_int),
        ('origin', Point),
        ('size', Point),
        ('indent', Point),
        ('periodicity', Periodicity),
        ('radius', Radius),
        ('counter', Counter),
        ('minimum_distance', c_longdouble),
        ('iterations', c_uint),
        ('exact_count', c_bool),
        ('order', c_short),
        ('heterogenous', c_bool),
    ]

    def toJSON(self):
        return {
            'dimension': self.dimension,
            'origin': self.origin(),
            'size': self.size(),
            'indent': self.indent(),
            'periodicity': self.periodicity(),
            'radius': self.radius(),
            'type': self.counter.type,
            'number': self.counter.number,
            'porosity': self.counter.porosity,
            'minimum_distance': self.minimum_distance,
            'iterations': self.iterations,
            'exact_count': self.exact_count,
            'order': self.order,
            'heterogenous': self.heterogenous,
        }

    @classmethod
    def fromJSON(cls, json):
        return cls(
            c_int(json['dimension']),
            Point(*json['origin']),
            Point(*json['size']),
            Point(*json['indent']),
            Periodicity(*json['periodicity']),
            Radius(*json['radius']),
            Counter(
                c_short(json['type']),
                c_uint(json['number']),
                c_longdouble(json['porosity'])
            ),
            c_longdouble(json['minimum_distance']),
            c_uint(json['iterations']),
            c_bool(json['exact_count']),
            c_short(json['order']),
            c_bool(json['heterogenous']),
        )

class Obstacle(Structure):
    _fields_ = [
        ('center', Point),
        ('radius', c_longdouble),
    ]

    def __repr__(self) -> str:
        return f'{self.center} ({self.radius})'
    
    def __call__(self) -> tuple:
        return self.center, self.radius



from random import random
from time import perf_counter
from math import sqrt

from . import __area


def __generate(size, periodicity, indent, porosityNumber, radius, minimumDistance, settings):
    indent_length = 0 if periodicity[0] else indent[0]
    indent_width = 0 if periodicity[1] else indent[1]
    indent_height = 0 if periodicity[2] else indent[2]

    indent_x = (indent_length + settings[3]) if not periodicity[0] else 0
    indent_y = (indent_width + settings[4]) if not periodicity[1] else 0
    indent_z = (indent_height + settings[5]) if not periodicity[2] else 0

    maximum_radius = min(min(settings[2] ** 2, settings[3] ** 2) - radius[0] - minimumDistance, radius[1])
    minimum_radius = maximum_radius if maximum_radius < radius[0] else radius[0]

    if settings[0] % 2 == 0:
        periodicity[0] = False
    if settings[1] % 2 == 0:
        periodicity[1] = False
    if settings[2] % 2 == 0:
        periodicity[2] = False

    spheres = set()
    spheres3d = [[[None for _ in range(settings[2])] for _ in range(settings[1])] for _ in range(settings[0])]

    total_volume = size[0] * size[1] * size[2]
    current_volume = float(total_volume)
    current_porosity = 0

    start = perf_counter()
    try:
        for i in range(settings[0]):
            if i == 0 and periodicity[0]:
                periodicity_x = settings[0] - 1
            else:
                periodicity_x = None
            for j in range(settings[1]):
                if j == 0 and periodicity[1]:
                    periodicity_y = settings[1] - 1
                else:
                    periodicity_y = None
                for k in range(settings[2]):
                    if k == 0 and periodicity[2]:
                        periodicity_z = settings[2] - 1
                    else:
                        periodicity_z = None
                    if (i + j + k) % 2 == 0:
                        if spheres3d[i][j][k] is None:
                            x, y, z = indent_x + settings[3] * i, indent_y + settings[4] * j, indent_z + settings[5] * k
                            max_r = maximum_radius
                            for n, m, t in ((1, 1, -1), (-1, -1, -1), (-1, 1, -1), (1, -1, -1), (1, 1, 1), (-1, -1, 1), (-1, 1, 1), (1, -1, 1)):
                                r, d, c = i + n, j + m, k + t
                                if 0 <= r < settings[0] and 0 <= d < settings[1] and 0 <= c < settings[2]:
                                    if spheres3d[r][d][c] is not None:
                                        max_rt = sqrt((spheres3d[r][d][c][0] - x) ** 2 + (spheres3d[r][d][c][1] - y) ** 2 + (spheres3d[r][d][c][2]) ** 2) - spheres3d[r][d][c][3] - minimumDistance
                                        max_r = min(max_rt, max_r)
                            r = minimum_radius + random() * (max_r - minimum_radius)
                            spheres3d[i][j][k] = (x, y, z, r)
                            if periodicity_x is not None:
                                if spheres3d[periodicity_x][j][k] is None:
                                    spheres3d[periodicity_x][j][k] = (indent_x + settings[3] * periodicity_x,
                                                                 y,
                                                                 z,
                                                                 spheres3d[i][j][k][3])
                            if periodicity_y is not None:
                                if spheres3d[i][periodicity_y][k] is None:
                                    spheres3d[i][periodicity_y][k] = (x,
                                                                 indent_y + settings[4] * periodicity_y,
                                                                 z,
                                                                 spheres3d[i][j][k][3])
                            if periodicity_z is not None:
                                if spheres3d[i][j][periodicity_z] is None:
                                    spheres3d[i][j][periodicity_z] = (x,
                                                                 y,
                                                                 indent_z + settings[5] * periodicity_z,
                                                                 spheres3d[i][j][k][3])
                        spheres.add(tuple(spheres3d[i][j][k]))

                        current_volume -= __area.__get_volume(size[0], size[1], size[2], *spheres3d[i][j][k], indent_length, indent_height, indent_width)
                        current_porosity = 100 * (current_volume / total_volume)

                        if len(spheres) >= porosityNumber:
                            raise KeyboardInterrupt
    except KeyboardInterrupt:
        ...

    end = perf_counter()

    return spheres, current_porosity, end - start



import gmsh
from math import sqrt

from . import __chess, __inline, __random


def __convert(geometry_data):
    geometry_data['size'] = float(geometry_data['size'][0]), float(geometry_data['size'][1]), float(geometry_data['size'][2])
    geometry_data['porosityNumber'] = float(geometry_data['porosityNumber'])
    geometry_data['radius'] = float(geometry_data['radius'][0]), float(geometry_data['radius'][1])
    geometry_data['minimumDistance'] = float(geometry_data['minimumDistance'])
    geometry_data['indent'] = float(geometry_data['indent'][0]), float(geometry_data['indent'][1]), float(geometry_data['indent'][2])


def __calculate_lines_indents(size_length, size_height, size_width, obstacles, periodicity_length, periodicity_height, periodicity_width):
    lines_x = round((obstacles * size_length / (size_height * size_width) ** (1 / 2)) ** (1 / 3))
    lines_y = round((obstacles * size_height / (size_length * size_width) ** (1 / 2)) ** (1 / 3))
    lines_z = round((obstacles * size_width / (size_length * size_height) ** (1 / 2)) ** (1 / 3))
    while lines_x * lines_y * (lines_z - 1) >= obstacles:
        lines_z -= 1
    while lines_x * lines_y * lines_z < obstacles:
        if lines_x * lines_y * (lines_z + 1) - obstacles > lines_z:
            lines_z += 1
        elif lines_x * (lines_y + 1) * lines_y - obstacles > lines_z:
            lines_y += 1
        else:
            lines_x += 1
    lines_indent_x = size_length / (lines_x + (1 if not periodicity_length else -1))
    lines_indent_y = size_height / (lines_y + (1 if not periodicity_height else -1))
    lines_indent_z = size_width / (lines_z + (1 if not periodicity_width else -1))
    return lines_x, lines_y, lines_z, lines_indent_x, lines_indent_y, lines_indent_z


def __generate(geometry_data):

    __convert(geometry_data)

    if geometry_data['order'] == 0:
        generator = __random

    elif geometry_data['order'] == 1:
        generator = __chess
        geometry_data['settings'] = __calculate_lines_indents(*geometry_data['size'], 2 * geometry_data['porosityNumber'] - 1, *geometry_data['periodicity'])

        del geometry_data['porosityNumberButton']

    elif geometry_data['order'] == 2:
        generator = __inline
        geometry_data['settings'] = __calculate_lines_indents(*geometry_data['size'], geometry_data['porosityNumber'], *geometry_data['periodicity'])

        del geometry_data['porosityNumberButton']

    else:
        raise TypeError

    del geometry_data['order']

    spheres, porosity, time = generator.__generate(**geometry_data)

    return spheres, porosity, time



from random import random
from time import perf_counter
from math import pi

from . import __area


def __check_for_intersection(x, y, z, r, spheres, minimal_distance):
    for _x, _y, _z, _r in spheres:
        if (_x - x) ** 2 + (_y - y) ** 2 + (_z - z) ** 2 <= (_r + r + minimal_distance) ** 2:
            return True
    return False


def __generate(size, periodicity, indent, porosityNumberButton, porosityNumber, radius, minimumDistance):
    print(size, periodicity, indent)
    indent_length = 0 if periodicity[0] else indent[0]
    indent_width = 0 if periodicity[1] else indent[1]
    indent_height = 0 if periodicity[2] else indent[2]

    spheres = set()

    total_volume = size[0] * size[1] * size[2]
    current_volume = float(total_volume)
    spheres_area = 0
    current_porosity = 100

    volume_calc = False
    running = True
    start = perf_counter()

    while running:
        if len(spheres) >= porosityNumber and porosityNumberButton == 1 or current_porosity <= porosityNumber and porosityNumberButton:
            running = False
        r = radius[0] + random() * (radius[1] - radius[0])

        if periodicity[0]:
            x = random.random() * size[0]
        else:
            x = indent_length + r + random() * (size[0] - 2 * r)

        if periodicity[1]:
            y = random() * size[1]
        else:
            y = indent_width + r + random() * (size[1] - 2 * r)

        if periodicity[2]:
            z = random() * size[2]
        else:
            z = indent_height + r + random() * (size[2] - 2 * r)

        if not __check_for_intersection(x, y, z, r, spheres, minimumDistance):
            need_x = 2 if (x - r < 0 or x + r > size[0] + 2 * indent_length) and periodicity[0] else 0
            need_y = 2 if (y - r < 0 or y + r > size[1] + 2 * indent_width) and periodicity[1] else 0
            need_z = 2 if (z - r < 0 or z + r > size[2] + 2 * indent_height) and periodicity[2] else 0
            if len(spheres) + need_x + need_y + need_z + (2 if need_x and need_y and need_z else 0) >= porosityNumber and porosityNumberButton == 1:
                continue
            copies = []
            if need_x:
                if x - r < 0:
                    _x = x + size[0]
                elif x + r > size[0] + indent_length:
                    _x = x - size[0]
                copies.append((_x, y, z))
            if need_y:
                if y - r < 0:
                    _y = y + size[1]
                elif y + r > size[1] + indent_width:
                    _y = y - size[1]
                copies.append((x, _y, z))
                if need_x:
                    copies.append((_x, _y, z))
            if need_z:
                if z - r < 0:
                    _z = z + size[2]
                elif z + r > size[2] + indent_height:
                    _z = z - size[2]
                copies.append((x, y, _z))
                if need_x:
                    copies.append((_x, y, _z))
                if need_y:
                    copies.append((x, _y, _z))
                if need_x and need_y:
                    copies.append((_x, _y, _z))
            for _x, _y, _z in copies:
                if __check_for_intersection(_x, _y, _z, r, spheres, minimumDistance):
                    break
            else:
                for _x, _y, _z in copies:
                    spheres.add((_x, _y, _z, r))
                    if volume_calc:
                        current_volume -= __area.__get_volume(size[0], size[1], size[2], _x, _y, _z, r, indent_length, indent_width, indent_height)
                spheres.add((x, y, z, r))
                if volume_calc:
                    current_volume -= __area.__get_volume(size[0], size[1], size[2], x, y, z, r, indent_length, indent_width, indent_height)
                else:
                    current_volume -= 4 / 3 * pi * r ** 3
                    spheres_area += 4 * pi * r ** 2
                current_porosity = 100 * (current_volume / total_volume)

    end = perf_counter()

    return spheres, current_porosity, end - start



try:
    import gmsh

    def __initialize_gmsh():
        __finalize_gmsh()
        gmsh.initialize()

    def __finalize_gmsh():
        if gmsh.isInitialized():
            gmsh.finalize()

    def __get_2d():
        return gmsh.model.occ.addRectangle, addDisk

    def __get_3d():
        return gmsh.model.occ.addBox, gmsh.model.occ.addSphere

    def addDisk(x, y, z, r):
        return gmsh.model.occ.addDisk(x, y, z, r, r)

    def __heterogenousBooleanOperations(domainDimTag, obstacleDimTags):
        obstacleDimTags = gmsh.model.occ.intersect(obstacleDimTags, domainDimTag, removeTool=False)[0]
        return gmsh.model.occ.cut(domainDimTag, obstacleDimTags, removeTool=False)[0], obstacleDimTags

    def __nonHeterogenousBooleanOperations(domainDimTag, obstacleDimTags):
        return gmsh.model.occ.cut(domainDimTag, obstacleDimTags)[0], []

    def __mark_2d_boundaries(size, eps=1e-3):
        for xmin, xmax, ymin, ymax, tag in (
            (0, 1, 1, 1, 1),  # top
            (1, 1, 0, 1, 2),  # right
            (0, 1, 0, 0, 3),  # bottom
            (0, 0, 0, 1, 4),  # left
        ):
            tags = gmsh.model.getEntitiesInBoundingBox(size[0]*xmin-eps, size[1]*ymin-eps, -eps, size[0]*xmax+eps, size[1]*ymax+eps, eps,dim=1)
            gmsh.model.addPhysicalGroup(1, tuple(tag for _, tag in tags), tag=tag)

        # topTags = gmsh.model.getEntitiesInBoundingBox(-eps, size[1]-eps, -eps, size[0]+eps, size[1]+eps, eps,dim=1)
        # gmsh.model.addPhysicalGroup(1, tuple(tag for _, tag in topTags), tag=1)

        # rightTags = gmsh.model.getEntitiesInBoundingBox(size[0]-eps, -eps, -eps, size[0]+eps, size[1]+eps, eps,dim=1)
        # gmsh.model.addPhysicalGroup(1, tuple(tag for _, tag in rightTags), tag=2)

        # bottomTags = gmsh.model.getEntitiesInBoundingBox(-eps, -eps, -eps, size[0]+eps, eps, eps,dim=1)
        # gmsh.model.addPhysicalGroup(1, tuple(tag for _, tag in bottomTags), tag=3)

        # leftTags = gmsh.model.getEntitiesInBoundingBox(-eps, -eps, -eps, eps, size[1]+eps, eps,dim=1)
        # gmsh.model.addPhysicalGroup(1, tuple(tag for _, tag in leftTags), tag=4)

        obstacleTags = gmsh.model.getEntitiesInBoundingBox(eps, eps, -eps, size[0]-eps, size[1]-eps, eps,dim=1)
        gmsh.model.addPhysicalGroup(1, tuple(tag for _, tag in obstacleTags), tag=5)

    def __mark_3d_boundaries(size, eps=1e-3):
        for xmin, xmax, ymin, ymax, zmin, zmax, tag in (
            (0, 1, 0, 1, 1, 1, 1),  # top
            (1, 1, 0, 1, 0, 1, 2),  # right
            (0, 1, 0, 1, 0, 0, 3),  # bottom
            (0, 0, 0, 1, 0, 1, 4),  # left
            (0, 1, 1, 1, 0, 1, 5),  # front
            (0, 1, 0, 0, 0, 1, 6),  # back
        ):
            tags = gmsh.model.getEntitiesInBoundingBox(size[0]*xmin-eps, size[1]*ymin-eps, size[2]*zmin-eps, size[0]*xmax+eps, size[1]*ymax+eps, size[2]*zmax+eps,dim=2)
            gmsh.model.addPhysicalGroup(2, tuple(tag for _, tag in tags), tag=tag)

        obstacleTags = gmsh.model.getEntitiesInBoundingBox(eps, eps, eps, size[0]-eps, size[1]-eps, size[2]-eps,dim=2)
        gmsh.model.addPhysicalGroup(2, tuple(tag for _, tag in obstacleTags), tag=7)

    def __set_options(meshSize, algorithm2d, algorithm3d):
        if algorithm2d is not None:
            gmsh.option.setNumber('Mesh.Algorithm', algorithm2d)
        if algorithm3d is not None:
            gmsh.option.setNumber('Mesh.Algorithm3D', algorithm3d)
        if meshSize is not None:
            gmsh.option.setNumber('Mesh.MeshSizeFactor', meshSize)
        gmsh.option.setNumber('Mesh.MshFileVersion', 2.2)

    def __generate_mesh(dim, refine=0):
        gmsh.model.mesh.generate(dim=dim)
        for _ in range(refine):
            gmsh.model.mesh.refine()
        if refine > 0:
            gmsh.model.mesh.generate(dim=dim)

    def __sync():
        gmsh.model.occ.synchronize()

    def __run_fltk():
        gmsh.fltk.initialize()
        gmsh.fltk.unlock()
        gmsh.fltk.run()
        gmsh.fltk.awake()
        gmsh.fltk.finalize()

    def __add_physical_group(dim, tags, tag):
        gmsh.model.addPhysicalGroup(dim, tags, tag=tag)

    def __get_nodes():
        nodes = gmsh.model.mesh.get_nodes()[:2]
        return tuple(nodes[0]), tuple(nodes[1])

    def __get_elements(dim, tag, meshElementType):
        elements = gmsh.model.mesh.getElements(dim=dim, tag=tag)
        index = tuple(elements[0]).index(meshElementType)
        return tuple(elements[1][index]), tuple(elements[2][index])

    def __load_file(fileName):
        gmsh.open(fileName)

except ModuleNotFoundError:
    pass



from setuptools import setup, find_packages

with open('README.md') as file:
    long_description = file.read()

setup(
    name='VPGen',
    version='1.0.0',
    author='Savvin Anton',
    author_email='anv.savvin@gmail.com',
    long_description_content_type='text/markdown',
    long_description=long_description,
    homepage='https://github.com/SavvinAnton/VPGen',
    license='GPL-3.0 license',
    description='Generate porosity medium',
    packages=find_packages(),
    install_requires=[],
    keywords=['python'],
    platforms=['Windows', 'Unix'],
    classifiers=[
        'Programming Language :: Python :: 3',
    ],
    include_package_data=True,
)



from time import perf_counter
from random import random
from math import sqrt

from . import __area


def __generate(size, periodicity, indent, porosityNumber, radius, minimumDistance, settings):
    indent_length = 0 if periodicity[0] else indent[0]
    indent_width = 0 if periodicity[1] else indent[1]

    indent_x = (indent_length + settings[2]) if not periodicity[0] else 0
    indent_y = (indent_width + settings[3]) if not periodicity[1] else 0

    if settings[0] % 2 == 0:
        periodicity[0] = False
    if settings[1] % 2 == 0:
        periodicity[1] = False

    maximum_radius = min(sqrt(settings[2] ** 2 + settings[3] ** 2) - radius[0] - minimumDistance, radius[1])
    minimum_radius = maximum_radius if maximum_radius < radius[0] else radius[0]

    disks = set()
    disks2d = [[None for _ in range(settings[1])] for _ in range(settings[0])]

    total_area = size[0] * size[1]
    current_area = float(total_area)
    current_porosity = 0

    start = perf_counter()
    try:
        for i in range(settings[0]):
            if i == 0 and periodicity[0]:
                periodicity_x = settings[0] - 1
            else:
                periodicity_x = None
            for j in range(settings[1]):
                if j == 0 and periodicity[1]:
                    periodicity_y = settings[1] - 1
                else:
                    periodicity_y = None
                if (i + j) % 2 == 0:
                    if disks2d[i][j] is None:
                        x, y = indent_x + settings[2] * i, indent_y + settings[3] * j
                        max_r = maximum_radius
                        for n, m in ((1, 1), (-1, -1), (-1, 1), (1, -1)):
                            r, d = i + n, j + m
                            if 0 <= r < settings[0] and 0 <= d < settings[1]:
                                if disks2d[r][d] is not None:
                                    max_rt = sqrt((disks2d[r][d][0] - x) ** 2 + (disks2d[r][d][1] - y) ** 2) - disks2d[r][d][2] - minimumDistance
                                    max_r = min(max_rt, max_r)
                        r = minimum_radius + random() * (max_r - minimum_radius)
                        disks2d[i][j] = (x, y, r)
                        if periodicity_x is not None:
                            if disks2d[periodicity_x][j] is None:
                                disks2d[periodicity_x][j] = (indent_x + settings[2] * periodicity_x,
                                                             y,
                                                             disks2d[i][j][2])
                        if periodicity_y is not None:
                            if disks2d[i][periodicity_y] is None:
                                disks2d[i][periodicity_y] = (x,
                                                             indent_y + settings[3] * periodicity_y,
                                                             disks2d[i][j][2])
                    disks.add((disks2d[i][j][0], disks2d[i][j][1], 0, disks2d[i][j][2]))

                    current_area -= __area.__getArea(size[0], size[1], *disks2d[i][j], indent_length, indent_width)
                    current_porosity = 100 * (current_area / total_area)

                if len(disks) >= porosityNumber:
                    raise KeyboardInterrupt
    except KeyboardInterrupt:
        ...

    end = perf_counter()

    return disks, current_porosity, end - start



try:
    from vtk import (
        vtkRenderWindow,
        vtkRenderer,
        vtkRenderWindowInteractor,
        vtkSphereSource,
        vtkPolyDataMapper,
        vtkActor,
        vtkInteractorStyleTrackballCamera,
        vtkPoints,
        vtkCellArray,
        vtkPolyData,
        vtkAxesActor,
        vtkOrientationMarkerWidget,
        vtkTextProperty,
    )
    USE_VTK = True
except ModuleNotFoundError:
    USE_VTK = False
try:
    from matplotlib import pyplot, patches
    USE_MATPLOTLIB = True
except ModuleNotFoundError:
    USE_MATPLOTLIB = False

from .Classes import Domain


def plot2d(domain: Domain, obstacles: tuple, savefig: str=None, indent: bool=True) -> None:
    '''
    Creates plot of obstacles (circles).

    domain: Domain - geometry
    obstacles: tuple - tuple of obstacles
    savefig: str - name of filename to save or will show if None
    indent: bool - show indent or not

    return - None
    '''

    if not USE_MATPLOTLIB:
        print('No matplotlib?')
        return

    if domain.dimension != 2:
        print('plot2d!')

    ax, fig = pyplot.subplots()

    for obstacle in obstacles:
        fig.add_patch(
            patches.Circle(
                (obstacle.center.x, obstacle.center.y), obstacle.radius
            )
        )
    fig.set_xlim(
        domain.origin.x + (domain.indent.x if indent else 0),
        domain.origin.x + domain.indent.x + (domain.indent.x if not indent else 0) + domain.size.x
    )
    fig.set_ylim(
        domain.origin.y + (domain.indent.y if indent else 0),
        domain.origin.y + domain.indent.y + (domain.indent.y if not indent else 0) + domain.size.y
    )

    if savefig is None:
        try:
            pyplot.show()
        except Exception as error:
            print('pyplot.show() error:', str(error))
            ax.savefig('_error.png')
    else:
        ax.savefig(savefig + '.png')


def plot3d(domain: Domain, obstacles:tuple) -> None:
    '''
    Creates plot of obstacles (spheres).

    domain: Domain - geometry
    obstacles: tuple - tuple of obstacles

    return - None
    '''

    if not USE_VTK:
        print('No vtk?')
        return

    if domain.dimension != 3:
        print('plot3d!')

    # TODO make using vtk

    ps = [
        [domain.origin.x, domain.origin.x + domain.size.x],
        [domain.origin.y, domain.origin.y + domain.size.y],
        [domain.origin.z, domain.origin.z + domain.size.z],
    ]

    points = vtkPoints()
    for i in range(2):
        for j in range(2):
            for k in range(2):
                points.InsertNextPoint(ps[0][i], ps[1][j], ps[2][k])

    lines = vtkCellArray()
    insertFace([1, 0, 2, 3], lines)
    insertFace([5, 4, 6, 7], lines)
    insertLine([0, 4], lines)
    insertLine([1, 5], lines)
    insertLine([2, 6], lines)
    insertLine([3, 7], lines)

    outline = vtkPolyData()
    outline.SetPoints(points)
    outline.SetLines(lines)

    outlineMapper = vtkPolyDataMapper()
    outlineMapper.SetInputData(outline)

    outlineActor = vtkActor()
    outlineActor.SetMapper(outlineMapper)

    outlineActor.GetProperty().SetColor(1, 1, 1)

    renderer = vtkRenderer()

    for obstacle in obstacles:
        renderer.AddActor(
            createSphereActor(obstacle.center, obstacle.radius)
        )
    renderer.AddActor(outlineActor)

    renderWindow = vtkRenderWindow()
    renderWindow.AddRenderer(renderer)
    
    interactor = vtkRenderWindowInteractor()
    interactor.SetInteractorStyle(vtkInteractorStyleTrackballCamera())
    interactor.SetRenderWindow(renderWindow)

    textProperty = vtkTextProperty()
    textProperty.SetColor(1, 1, 1)
    textProperty.SetBold(False)
    textProperty.SetItalic(False)
    textProperty.SetShadow(False)

    axesActor = vtkAxesActor()
    axesActor.GetXAxisCaptionActor2D().SetCaptionTextProperty(textProperty)
    axesActor.GetYAxisCaptionActor2D().SetCaptionTextProperty(textProperty)
    axesActor.GetZAxisCaptionActor2D().SetCaptionTextProperty(textProperty)
    
    orientationMarker = vtkOrientationMarkerWidget()
    orientationMarker.SetOrientationMarker(axesActor)
    orientationMarker.SetInteractor(interactor)
    orientationMarker.EnabledOn()
    orientationMarker.InteractiveOn()
    orientationMarker.SetOutlineColor(1, 1, 1)

    renderer.ResetCamera()

    interactor.Initialize()
    interactor.Start()


def createSphereActor(center, radius):
    source = vtkSphereSource()
    source.SetCenter(*center())
    source.SetRadius(radius)
    source.Update()

    mapper = vtkPolyDataMapper()
    mapper.SetInputData(source.GetOutput())

    actor = vtkActor()
    actor.SetMapper(mapper)

    actor.GetProperty().SetColor(1, 1, 1)

    return actor


def insertFace(ids, lines):
    lines.InsertNextCell(5)
    for id in ids + [ids[0]]:
        lines.InsertCellPoint(id)


def insertLine(ids, lines):
    lines.InsertNextCell(2)
    for id in ids:
        lines.InsertCellPoint(id)



from ctypes import POINTER
from time import perf_counter
from os.path import abspath, dirname
from platform import system

from .IO import loadGeometry, loadLibrary, CALLBACK_TYPE
from .Classes import Domain, Obstacle, Point
from .Old import generateGeometry


def createDomain(data: dict) -> Domain:
    '''
    Creates 'Domain' from 'dict'.

    data: dict - dict with domain settings

    return - Domain
    '''
    # return Domain(
    #         c_int(data['dimension']),
    #         Point(*data['origin']),
    #         Point(*data['size']),
    #         Point(*data['indent']),
    #         Periodicity(*data['periodicity']),
    #         Radius(
    #             c_longdouble(min(data['radius'])),
    #             c_longdouble(max(data['radius']))
    #         ),
    #         Counter(
    #             c_short(data['type']),
    #             c_uint(data['number']),
    #             c_longdouble(data['porosity'])
    #         ),
    #         c_longdouble(data['minimum_distance']),
    #         c_uint(data['iterations']),
    #         c_bool(data['exact_count']),
    #         c_short(data['order']),
    #         c_bool(data['heterogenous']),
    #     )
    return Domain.fromJSON(data)


def generateFromFile(filename: str) -> tuple:
    '''
    Generates from file with domain settings and return tuple of obstacles, number of obstacles, porosity, generating time.

    filename: str - path to .json file with domain settings

    return - tuple
    '''
    domain = createDomain(loadGeometry(filename))
    return generate(domain)


def generate(domain: Domain, callback_function=lambda *_: None) -> tuple:
    '''
    Generates from domain and return tuple of obstacles, number of obstacles, porosity, generating time.

    domain: Domain - geometry
    callback_function: callable - function that called every iteration, must be (int, float) -> None

    return - tuple
    '''
    try:
        lib = loadLibrary(LIB_PATH)
        LIBRARY_LOADED = True
    except Exception as error:
        print(f'[Error] {error}')
        LIBRARY_LOADED = False

    if LIBRARY_LOADED and domain.order == 0:
        callback_function = CALLBACK_TYPE(callback_function)

        start = perf_counter()
        obstacles_number, porosity = lib.generate(domain, callback_function).contents
        end = perf_counter()

        obstacles_number = int(obstacles_number)
        lib.getObstacles.restype = POINTER(Obstacle * obstacles_number)
        obstacles = lib.getObstacles().contents
        return obstacles, obstacles_number, porosity, end - start
    else:
        geometry_data = {
            'size': list(domain.size()),
            'heterogenous': domain.heterogenous,
            'dimension': domain.dimension,
            'periodicity': list(domain.periodicity()),
            'indent': list(domain.indent()),
            'radius': list(domain.radius()),
            'porosityNumberButton': 1-domain.counter.type,
            'porosityNumber': domain.counter.porosity if domain.counter.type == 1 and domain.order == 0 else domain.counter.number,
            'minimumDistance': domain.minimum_distance,
            'order': domain.order,
        }
        obstacles, porosity, time, heterogenous, size, dimension = generateGeometry(geometry_data)
        obstacles = list(obstacles)
        for i, obstacle in enumerate(obstacles):
            obstacles[i] = Obstacle(Point(*obstacle[:3]), obstacle[3])
    return obstacles, len(obstacles), porosity, time


LIBRARY_DIRECTORY = dirname(abspath(__file__))
LIB_PATH = LIBRARY_DIRECTORY
SYSTEM = system()
if SYSTEM == 'Linux':
    LIB_PATH += '/VPGen.so'
elif SYSTEM == 'Windows':
    LIB_PATH += '/VPGen.dll'
else:
    raise OSError('Only Linux/Windows')



try:
    import gmsh
    USE_GMSH = True
except ModuleNotFoundError:
    USE_GMSH = False

from . import __core, __pgen, __vgen
# import __core, __pgen, __vgen


def generateGeometry(geometry_data):
    dimension = geometry_data['dimension']
    heterogenous = geometry_data['heterogenous']

    del geometry_data['dimension']
    del geometry_data['heterogenous']

    if dimension == 2:
        obstacles, porosity, time = __pgen.__generate(geometry_data)
        size = (
            geometry_data['size'][0] + (0 if geometry_data['periodicity'][0] else (2 * geometry_data['indent'][0])),
            geometry_data['size'][1] + (0 if geometry_data['periodicity'][1] else (2 * geometry_data['indent'][1]))
        )

    elif dimension == 3:
        obstacles, porosity, time = __vgen.__generate(geometry_data)
        size = (
            geometry_data['size'][0] + (0 if geometry_data['periodicity'][0] else (2 * geometry_data['indent'][0])),
            geometry_data['size'][1] + (0 if geometry_data['periodicity'][1] else (2 * geometry_data['indent'][1])),
            geometry_data['size'][2] + (0 if geometry_data['periodicity'][2] else (2 * geometry_data['indent'][2]))
        )

    else:
        raise ValueError('[generateGeometry] Wrong dimension')

    return obstacles, porosity, time, heterogenous, size, dimension


def generateGeometrySkeleton(obstacles, heterogenous, size, dimension, fltk=False):
    if not USE_GMSH:
        raise ImportError('GMSH is not found!')

    __core.__initialize_gmsh()

    if dimension == 2:
        domainType, obstacleType = __core.__get_2d()
    elif dimension == 3:
        domainType, obstacleType = __core.__get_3d()
    else:
        raise ValueError('[generateGeometrySkeleton] Wrong dimension')

    booleanOperations = __core.__heterogenousBooleanOperations if heterogenous else __core.__nonHeterogenousBooleanOperations

    domainDimTag = ((dimension, domainType(0, 0, 0, *size)),)
    obstacleDimTags = tuple((dimension, obstacleType(*obstacle)) for obstacle in obstacles)

    domainDimTag, obstacleDimTags = booleanOperations(domainDimTag, obstacleDimTags)

    __core.__sync()

    #   if dim == 0 --> 3 algorithm2d
    # elif dim == 1 --> 2 algorithm2d
    # else 'what?'

    __core.__set_options(None, 3-dimension, None)
    __core.__generate_mesh(dim=2, refine=dimension)

    if fltk:
        __core.__run_fltk()

    nodeTags, nodeCoords = __core.__get_nodes()

    domains = []
    if dimension == 2:
        elementTags, elementNodeTags = __core.__get_elements(2, domainDimTag[0][1], 2)
        domains.append((elementTags, elementNodeTags, (0, 0, 1)))
        for _, tag in obstacleDimTags:
            elementTags, elementNodeTags = __core.__get_elements(2, tag, 2)
            domains.append((elementTags, elementNodeTags, (1, 0, 0)))

    elif dimension == 3:
        for _, tag in gmsh.model.getEntities(dim=2):
            if gmsh.model.getType(2, tag) == 'Plane':
                elementTags, elementNodeTags = __core.__get_elements(2, tag, 2)
                domains.append((elementTags, elementNodeTags, (0, 0, 1)))
            color = (1, 0, 0) if heterogenous else (0, 0, 1)
            if gmsh.model.getType(2, tag) == 'Sphere':
                elementTags, elementNodeTags = __core.__get_elements(2, tag, 2)
                domains.append((elementTags, elementNodeTags, color))
        # for _, tag in obstacleDimTags:
        #     elementTags, elementNodeTags = __core.__get_elements(2, tag, 2)
        #     domains.append((elementTags, elementNodeTags, (1, 0, 0)))

    else:
        ...

    __core.__finalize_gmsh()

    return (nodeTags, nodeCoords), domains


def generateMesh(obstacles, heterogenous, size, dimension, folder, meshSize=1, algorithm2d=2, algorithm3d=10, fltk=False):
    if not USE_GMSH:
        raise ImportError('GMSH is not found!')

    __core.__initialize_gmsh()

    if dimension == 2:
        domainType, obstacleType = __core.__get_2d()
    elif dimension == 3:
        domainType, obstacleType = __core.__get_3d()
    else:
        raise ValueError('[generateMesh] Wrong dimension')

    booleanOperations = __core.__heterogenousBooleanOperations if heterogenous else __core.__nonHeterogenousBooleanOperations

    domainDimTag = ((dimension, domainType(0, 0, 0, *size)),)
    obstacleDimTags = tuple((dimension, obstacleType(*obstacle)) for obstacle in obstacles)

    domainDimTag, obstacleDimTags = booleanOperations(domainDimTag, obstacleDimTags)

    __core.__sync()

    __core.__add_physical_group(dim=dimension, tags=(domainDimTag[0][1],), tag=1)
    if obstacleDimTags:
        __core.__add_physical_group(dim=dimension, tags=tuple(tag for _, tag in obstacleDimTags), tag=2)
    if dimension == 2:
        __core.__mark_2d_boundaries(size)
    elif dimension == 3:
        __core.__mark_3d_boundaries(size)

    __core.__set_options(meshSize, algorithm2d, algorithm3d)
    __core.__generate_mesh(dim=dimension, refine=0)

    if fltk:
        __core.__run_fltk()

    gmsh.write(folder + '/temp.msh')

    meshElementType = 2 if dimension == 2 else 4

    nodeTags, nodeCoords = __core.__get_nodes()
    elementTags, elementNodeTags = __core.__get_elements(-1, -1, meshElementType)

    __core.__finalize_gmsh()

    return (nodeTags, nodeCoords), (elementTags, elementNodeTags)


def loadMesh(fileName, dimension):
    if not USE_GMSH:
        raise ImportError('GMSH is not found!')

    __core.__initialize_gmsh()

    __core.__load_file(fileName)

    meshElementType = 2 if dimension == 2 else 4

    nodeTags, nodeCoords = __core.__get_nodes()
    elementTags, elementNodeTags = __core.__get_elements(-1, -1, meshElementType)

    __core.__finalize_gmsh()

    return (nodeTags, nodeCoords), (elementTags, elementNodeTags)
